﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Connect4
{
    static class AI
    {
        public const int profondeur = 4;

        //Ici se trouve les points pour chaccun des coups effectués, les points sont modifié par *-1 lorsque l'adversaire qui joue
        public const int pointsVictoire = 5000; //point lorsque l'on a une victoire, donc un très bon coups
        public const int pointsBlocage = 3000; //point lorsque l'on bloque le coups gagnant de l'adversaire
        public const int pointsEnleveOpportunite = -1000; //lorsque l'adversaire peut bloquer une victoire en jouant directement au dessus de la case où on joue
        public const int pointsDonneOpportunite = -1000;//lorsque l'aversaire joue sur la case au dessus de la notre et gagne, c'est un mauvais coup
        public const int pointsAdjacant = 100; //point pour le placement de plusieurs jetons en ligne, colonne ou diagonale, multiplié par le nombre de jeton adjaçant
        public const int MaximumPointsAdjacant = 250; //le maximum de point que l'on donne pour le placement adjacant
        //Règle à suivre : |pointsVictoire| > |pointsBlocage| > |pointsEnleveOpportunite| >= |pointsDonneOpportunite| >= |MaximumPointsadjacant| > |pointsadjacant|

        public const int facteurEnfant = 3; //facteur qui donne de l'aide pour les enfants
        public const int facteurParent = 4; //facteur qui donne de l'aide pour les parents
        public const int facteurProfondeurAdd = 1;

        public const int facteurImmediat = 500000; //facteur pour les enfants immédiats, très grand parce que ça veut dire bloquer l'adversaire le plus tôt possible
        public const int facteurPasEnfant = 2; //facteur qui donne de l'aide pour les parents qui n'ont pas d'enfant, donc défavorisé par rapport à ceux ayant des enfants
        

        /// <summary>
        /// Fonction principale de l'AI, on l'appel et retourne la position à jouer
        /// </summary>
        /// <param name="p_Board">Le board du jeu, contenant des y, r et n</param>
        /// <param name="p_Profondeur">La profondeur utilisé, normalement il s'agit de la variable profondeur définie dans nos constantes plus haut</param>
        /// <param name="p_Player">Le player qui joue actuellement</param>
        /// <returns>Une position où on doit jouer</returns>
        static public Position Choisir(char[,] p_Board, int p_Profondeur, Player p_Player)
        {
            List<ScorePosition> nosDecisions = new List<ScorePosition>();//contient une liste de score et de position
            List<Position> nosEnfants = TrouverEnfants(p_Board);//on recherche nos enfants

            bool couleurJaune = !p_Player.IsRed; //on veut savoir si on place un jaune ou un rouge

            foreach (Position pos in nosEnfants)//pour tous nos enfants
            {
                char[,] p_BoardCopy = CopierTableauChar(p_Board);//on crée une copie du tableau

                //on modifie la valeur où on teste nos enfants, c'est-à-dire, un y ou un r pour chaque colonne
                if (couleurJaune == true)
                {
                    p_BoardCopy[pos.colonne, pos.ligne] = 'y';
                }
                else
                {
                    p_BoardCopy[pos.colonne, pos.ligne] = 'r';
                }

                //on appel à chacun de nos enfants pour qu'ils testent leurs enfants
                //on trouve le score de nos enfants que l'on majore avec des points de pion adjacant
                int score = TrouverScoreEnfants(p_BoardCopy, p_Profondeur - 1, !couleurJaune, p_Player, pos) + pionAdjacant(p_BoardCopy, pos);
                nosDecisions.Add(new ScorePosition(score, pos));//on crée un score position avec la position et le score trouvé
            }


            // ce bloc de code sélectionne le meilleur, ou les meilleurs score, s'il y a plus qu'un score égal
            List<ScorePosition> meilleursChoix = new List<ScorePosition>(); //une liste de meilleurs choix que l'on trouvera
            nosDecisions.Sort(ScorePosition.ComparaisonScore); //on trie notre tableau de score
            int meilleurScore = nosDecisions[nosDecisions.Count - 1].score; //le dernier élément est donc forcément notre meilleur score
            foreach (ScorePosition scorePos in nosDecisions)
            {
                if (scorePos.score == meilleurScore)//si notre score est égale au meilleur score, on l'ajoute au tableau
                {
                    meilleursChoix.Add(scorePos);
                }
            }

            //on prend parmi la liste des meilleurs scores, un au hasard
            System.Random hasard = new Random((int)DateTime.Now.Ticks);
            int index = hasard.Next(0, meilleursChoix.Count);

            return meilleursChoix[index].position;
        }

        /// <summary>
        /// Détermine le meilleur choix parmi les enfants, selon une recherche en profondeur
        /// </summary>
        /// <param name="p_Board">L'état du tableau à l'endroit où l'on est rendu dans l'arbre</param>
        /// <param name="p_Profondeur">La profondeur où on est rendu dans l'arbre</param>
        /// <param name="couleurJaune">Si on est jaune ou pas</param>
        /// <param name="p_Player">Le joueur à qui c'est le tour de jouer</param>
        /// <param name="position">Une position où notre parent à modifier ma valeur</param>
        /// <returns>Un int qui contiendra la moyenne pondéré des scores de mes enfants ou qui contiendra uniquement mon score</returns>
        static private int TrouverScoreEnfants(char[,] p_Board, int p_Profondeur, bool couleurJaune, Player p_Player, Position position)
        {
            //un facteur multiplicatif pour -1 pour coup de l'ennemie et 1 pour nos coups, donc on estime si l'aversaire joue de son mieux
            int facteurMult = -1;
            if ((p_Player.IsRed && !couleurJaune) || (!p_Player.IsRed && couleurJaune))//si je joue (le joueur est jaune et ce noeud jaune ou le joueur est red et ce noeud red) 
            {
                facteurMult = 1;
            }

            int facteurProfondeur = p_Profondeur + facteurProfondeurAdd;//facteur en fonction de la pronfondeur dans le tableau pour aider le score du parent

            //pour savoir si c'est gagnant un gagnant immédiatement avec le premier fils ou pas
            int facteurGagnant = 1;
            if (p_Profondeur == (profondeur - 1))
            {
                facteurGagnant = facteurImmediat;
            }

            if (Engine.CheckWinState(p_Board[position.colonne, position.ligne], position.ligne, position.colonne, p_Board))//on vérifie que l'on gagne
            {
                //note on multiplie par p_profondeur parce qu'il n'y a pas d'enfant, donc on augmente la valeur parce que les enfants serait en théorie aussi gagnant
                return pointsVictoire * facteurProfondeur * facteurGagnant;
            }

            if (p_Profondeur >= 1)//si on regarde encore les enfants
            {
                List<Position> nosEnfants = TrouverEnfants(p_Board);//on recherche la liste des enfants
                List<int> nosScores = new List<int>();//contiendra la liste des scores de nos enfants

                foreach (Position pos in nosEnfants)//pour tous nos enfants
                {
                    char[,] p_BoardCopy = CopierTableauChar(p_Board);//on crée une copie du tableau

                    //on modifie la valeur oÙ on teste nos enfants
                    if (couleurJaune == true)
                    {
                        p_BoardCopy[pos.colonne, pos.ligne] = 'y';
                    }
                    else
                    {
                        p_BoardCopy[pos.colonne, pos.ligne] = 'r';
                    }

                    nosScores.Add(TrouverScoreEnfants(p_BoardCopy, p_Profondeur - 1, !couleurJaune, p_Player, pos));//on appelle l'enfant pour savoir son score
                }

                if (nosScores.Count == 0)//si la partie est pleine, donc que l'on a aucun enfant, on prend que notre score
                {
                    return CalculerScore(p_Board, position) * facteurProfondeur * facteurPasEnfant;
                }
                else
                {
                    //si on a des enfants on prend le meilleur plus notre score à nous
                    int monscore = (CalculerScore(p_Board, position)) * facteurProfondeur * facteurParent;
                    int score = (nosScores.Sum() / nosScores.Count) * facteurEnfant * facteurMult;
                    return score + monscore;
                }
            }
            else
            {
                //on n'a pas d'enfant parce qu'on est au plus profond de nos requêtes, on retourne que notre score, ici le facteur mult vaut toujours 1.
                return CalculerScore(p_Board, position) * facteurMult;
            }
            
        }


        /// <summary>
        /// Sert à trouver la liste de nos enfants possibles
        /// </summary>
        /// <param name="p_Board">Le board du jeu actuellement dans lequel on veut trouver les enfants</param>
        /// <returns>Une liste de position où on doit jouer et donc où sont nos enfants</returns>
        static private List<Position> TrouverEnfants(char[,] p_Board)
        {
            List<Position> nosPositions = new List<Position>();

            for (int i = 0; i < 7; i++)
            {
                int j = 5;
                bool stop = false;
                while (!stop && j >= 0)
                {
                    if (p_Board[i, j] == 'n')
                    {
                        stop = true;
                        nosPositions.Add(new Position(i, j));
                    }
                    j--;
                }
            }

            return nosPositions;
        }


        /// <summary>
        /// Calcule notre score 
        /// </summary>
        /// <param name="p_Board">Le board du jeu actuellement dans lequel on veut calculer le score d'un enfant</param>
        /// <param name="pos">La position de jeton auquel on veut calculer un score selon le p_Board</param>
        /// <returns>Un int contenant le score du jeton à la position pos selon le board p_Board</returns>
        static private int CalculerScore(char[,] p_Board, Position pos)
        {
            int score = 0;

            if (onBloque(p_Board, pos))//Si on bloque
            {
                score += pointsBlocage;
            }

            if (onEnleveUneOpportunite(p_Board, pos))//Si on s'enlève une opportunité
            {
                score += pointsEnleveOpportunite;
            }

            if (onDonneUneOpportunite(p_Board, pos))//Si on donne une opportunité à l'adversaire
            {
                score += pointsDonneOpportunite;
            }

            return score;
        }


        /// <summary>
        /// Le calcule le score en fonction des pions adjaçants
        /// </summary>
        /// <param name="p_Board">Le board du jeu actuellement dans lequel on veut calculer le score d'un enfant</param>
        /// <param name="pos">La position de jeton auquel on veut calculer un score selon le p_Board</param>
        /// <returns>Le score des pions adjaçant, qui vaut au maximum la valeur de MaximumPointsAdjacant</returns>
        static private int pionAdjacant(char[,] p_Board, Position pos)
        {
            int nombreAdjacant = 0;

            //pour la colonne suivant s'il y en a une
            if (pos.colonne + 1 < 7)
            {
                if (pos.ligne + 1 < 6 && p_Board[pos.colonne + 1, pos.ligne + 1] == p_Board[pos.colonne, pos.ligne])//pour la ligne au dessus
                {
                    nombreAdjacant++;
                }

                if (p_Board[pos.colonne + 1, pos.ligne] == p_Board[pos.colonne, pos.ligne])//pour la ligne au centre, directement à droite
                {
                    nombreAdjacant++;
                }

                if (pos.ligne - 1 >= 0 && p_Board[pos.colonne + 1, pos.ligne - 1] == p_Board[pos.colonne, pos.ligne])//pour la ligne au dessous
                {
                    nombreAdjacant++;
                }
            }

            //pour la colonne précédente s'il y en a une
            if (pos.colonne - 1 >= 0)
            {
                if (pos.ligne + 1 < 6 && p_Board[pos.colonne - 1, pos.ligne + 1] == p_Board[pos.colonne, pos.ligne])//pour la ligne au dessus
                {
                    nombreAdjacant++;
                }

                if (p_Board[pos.colonne - 1, pos.ligne] == p_Board[pos.colonne, pos.ligne])//pour la ligne au centre, directement à droite
                {
                    nombreAdjacant++;
                }

                if (pos.ligne - 1 >= 0 && p_Board[pos.colonne - 1, pos.ligne - 1] == p_Board[pos.colonne, pos.ligne])//pour la ligne au dessous
                {
                    nombreAdjacant++;
                }
            }

            //Pour la même colonne que l'on se trouve actuellement
            if (pos.ligne + 1 < 6 && p_Board[pos.colonne, pos.ligne + 1] == p_Board[pos.colonne, pos.ligne])
            {
                nombreAdjacant++;
            }

            if (pos.ligne - 1 >= 0 && p_Board[pos.colonne, pos.ligne - 1] == p_Board[pos.colonne, pos.ligne])
            {
                nombreAdjacant++;
            }


            //on retourne le minimum entre le nombre de point multiplié par nos enfants et le maximum de point que l'on accorde au adjaçant
            return Math.Min(nombreAdjacant * pointsAdjacant, MaximumPointsAdjacant);
        }


        /// <summary>
        /// Vérifie s'il en jouant à cette position on bloque
        /// </summary>
        /// <param name="p_Board">Le board du jeu actuellement dans lequel on veut calculer le score d'un enfant</param>
        /// <param name="pos">La position de jeton auquel on veut calculer un score selon le p_Board</param>
        /// <returns>true si on bloque, false sinon</returns>
        static private bool onBloque(char[,] p_Board, Position pos)
        {
            char[,] board = CopierTableauChar(p_Board);

            if (board[pos.colonne, pos.ligne] == 'y')
            {
                board[pos.colonne, pos.ligne] = 'r';
            }
            else
            {
                board[pos.colonne, pos.ligne] = 'y';
            }

            // On teste si cette position est alors gagnante, si oui cela signifie que l'on bloque
            if (Engine.CheckWinState(board[pos.colonne, pos.ligne], pos.ligne, pos.colonne, board))
            {
                return true;
            }

            return false;
        }


        /// <summary>
        /// On vérifie si l'on donne pas une opportunité
        /// </summary>
        /// <param name="p_Board">Le board du jeu actuellement dans lequel on veut calculer le score d'un enfant</param>
        /// <param name="pos">La position de jeton auquel on veut calculer un score selon le p_Board</param>
        /// <returns>true si on donne une opportunite, false sinon</returns>
        static private bool onDonneUneOpportunite(char[,] p_Board, Position pos)
        {
            int j = pos.ligne - 1;

            char[,] board = CopierTableauChar(p_Board);

            if (j < 0)
            {
                return false;
            }

            if (board[pos.colonne, pos.ligne] == 'y')
            {
                board[pos.colonne, j] = 'r';
            }
            else
            {
                board[pos.colonne, j] = 'y';
            }

            // On teste si cette position est alors gagnante, si oui cela signifie que l'on bloque
            if (Engine.CheckWinState(board[pos.colonne, j], j, pos.colonne, board))
            {
                return true;
            }

            return false;
        }


        /// <summary>
        /// On vérifie si l'on s'enlève pas une opportunité
        /// </summary>
        /// <param name="p_Board">Le board du jeu actuellement dans lequel on veut calculer le score d'un enfant</param>
        /// <param name="pos">La position de jeton auquel on veut calculer un score selon le p_Board</param>
        /// <returns>true si on s'enlève une opportunite, false sinon</returns>
        static private bool onEnleveUneOpportunite(char[,] p_Board, Position pos)
        {
            int j = pos.ligne - 1;

            char[,] board = CopierTableauChar(p_Board);

            if (j < 0)
            {
                return false;
            }

            if (board[pos.colonne, pos.ligne] == 'y')
            {
                board[pos.colonne, j] = 'y';
                board[pos.colonne, pos.ligne] = 'r';
            }
            else
            {
                board[pos.colonne, j] = 'r';
                board[pos.colonne, pos.ligne] = 'y';
            }

            // On teste si cette position est alors gagnante, si oui cela signifie que l'on bloque
            if (Engine.CheckWinState(board[pos.colonne, j], j, pos.colonne, board))
            {
                return true;
            }

            return false;
        }

        /// <summary>
        /// Copier un tableau de char[,]
        /// </summary>
        /// <param name="p_Source">Le tableau à copier</param>
        /// <returns>Le tableau copier</returns>
        static private char[,] CopierTableauChar(char[,] p_Source)
        {
            char[,] destination = new char[7, 6];

            for (int i = 0; i < 7; i++)
            {
                for(int j = 0; j < 6; j++)
                {
                    destination[i, j] = p_Source[i, j];
                }
            }

            return destination;
        }
    }


}
