﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;

namespace Connect4
{
    public partial class FrmPrinc : Form
    {
        //Dans le tableau board, le contenu est le suivant :
        //r = red
        //y = yellow
        //n = none

        public bool playerWon = false;
        public Player player1;
        public Player player2;
        public Player currentPlayer;
        public int tokensOnBoard = 0;

        public FrmPrinc()
        {
            InitializeComponent();
        }

        private void FrmPrinc_Load(object sender, EventArgs e)
        {
            PlayersSettings();
            txtMessages.Text = "";
            Engine.board = new char[7, 6];
            for (int i = 0; i < 7; i++)
            {
                for (int j = 0; j < 6; j++)
                {
                    Engine.board[i, j] = 'n';
                }
            }

            int decalageX = 94;
            int decalageY = 83;

            for (int i = 0; i < 7; i++)
            {
                for (int j = 0; j < 6; j++)
                {
                    PictureBox pic = new PictureBox();
                    if (j == 0)
                    {
                        pic.Location = new Point((decalageX * i) + i / 3, 0);
                        pic.Size = new Size(94, 88);
                    }
                    else if (j == 5)
                    {
                        pic.Location = new Point((decalageX * i) + i / 3, (decalageY * j - 3));
                        pic.Size = new Size(94, 88);
                    }
                    else if (j == 4)
                    {
                        pic.Location = new Point((decalageX * i) + i / 3, (decalageY * j - 2));
                        pic.Size = new Size(94, 83);
                    }
                    else
                    {
                        pic.Location = new Point((decalageX * i) + i / 3, (decalageY * j));
                        pic.Size = new Size(94, 83);
                    }

                    pic.Name = "picToken" + i.ToString() + j.ToString();
                    pic.BackColor = Color.Transparent;
                    picBoard.Controls.Add(pic);
                }
            }

            for (int i = 0; i < 7; i++)
            {
                TransparentPanel pnl = new TransparentPanel();
                pnl.Location = new Point(33 + (decalageX * i), 87);
                pnl.Size = new Size(94, 508);
                pnl.Name = "pnlCol" + i.ToString();
                this.Controls.Add(pnl);
                pnl.BringToFront();
                pnl.MouseEnter += new EventHandler(MouseEnterTransPnl);
                pnl.MouseLeave += new EventHandler(MouseLeaveTransPnl);
                if (Engine.gameType != PlayersType.ComputerComputer)
                {
                    pnl.MouseClick += new MouseEventHandler(MouseClickTransPnl);
                }
                else
                {
                    timerCompVSComp.Start();
                }
            }
        }

        private void PlayersSettings()
        {
            FrmPlayersChoice playersChoice = new FrmPlayersChoice();
            playersChoice.ShowDialog();
            Engine.gameType = playersChoice.ChosenType;

            switch (playersChoice.ChosenType)
            {
                case (PlayersType.HumanHuman):
                    {
                        player1 = new Player(PlayerType.Human, true, 1);
                        player2 = new Player(PlayerType.Human, false, 2);
                        lblPlayersType.Text = "Connect4 - Humain VS Humain";
                        break;
                    }
                case (PlayersType.HumanComputer):
                    {
                        player1 = new Player(PlayerType.Human, true, 1);
                        player2 = new Player(PlayerType.Computer, false, 2);
                        lblPlayersType.Text = "Connect4 - Humain VS Ordinateur";
                        break;
                    }
                case (PlayersType.ComputerComputer):
                    {
                        player1 = new Player(PlayerType.Computer, true, 1);
                        player2 = new Player(PlayerType.Computer, false, 2);
                        lblPlayersType.Text = "Connect4 - Ordinateur VS Ordinateur";
                        break;
                    }
            }

            currentPlayer = player1;
        }

        private void MouseEnterTransPnl(object sender, EventArgs e)
        {
            TransparentPanel pnl = (TransparentPanel)sender;
            pnl.transparent = false;
            pnl.Refresh();
            GC.Collect();
        }

        private void MouseLeaveTransPnl(object sender, EventArgs e)
        {
            TransparentPanel pnl = (TransparentPanel)sender;
            pnl.transparent = true;
            pnl.Hide();
            pnl.Refresh();
            pnl.Show();
            GC.Collect();
        }

        private void MouseClickTransPnl(object sender, EventArgs e)
        {
            if (currentPlayer.Type == PlayerType.Human)
            {
                Panel pnl = (Panel)sender;
                int col = (int)Char.GetNumericValue(pnl.Name[6]);
                bool ok = false;
                int result = -1;

                for (int i = 5; i >= 0 && !ok; i--)
                {
                    if (Engine.board[col, i] == 'n')
                    {
                        ok = true;
                        result = i;
                    }
                }

                if (ok && result > -1)
                {
                    MettreJeton(new Position(col, result));

                    if (Engine.gameType == PlayersType.HumanComputer && !playerWon)
                    {
                        Position posDeJeu = AI.Choisir(Engine.board, AI.profondeur, currentPlayer);
                        MettreJeton(posDeJeu);
                    }
                    else if (playerWon)
                    {
                        playerWon = false;
                    }
                }
                else
                {
                    txtMessages.Text = "La colonne est remplie! Vous ne pouvez pas y ajouter votre jeton! Choisissez une autre colonne!";
                    timerMessages.Start();
                }
            }
        }

        private void ResetBoard()
        {
            foreach (Control pic in picBoard.Controls)
            {
                PictureBox picBox = (PictureBox)pic;
                picBox.Image = null;
            }
            Engine.clearBoard();
            Engine.turn = 1;
            tokensOnBoard = 0;
            lblTurn.Text = Engine.turn.ToString();
            lblCurrentPlayer.Text = "1 - Rouge";
            currentPlayer = player1;
        }

        private void timerMessages_Tick(object sender, EventArgs e)
        {
            txtMessages.Text = "";
            timerMessages.Stop();
        }

        private void btnQuit_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        private void btnRestart_Click(object sender, EventArgs e)
        {
            ResetBoard();
        }

        private void btnSave_Click(object sender, EventArgs e)
        {
            SaveFileDialog sfd = new SaveFileDialog();
            sfd.Filter = "Text files|*.txt";
            DialogResult result = sfd.ShowDialog();

            if (result == DialogResult.OK)
            {
                using (StreamWriter outfile = new StreamWriter(sfd.FileName))
                {
                    string line = "";
                    for (int j = 0; j < 6; j++)
                    {
                        for (int i = 0; i < 7; i++)
                        {
                            line += Engine.board[i, j];
                        }
                        outfile.WriteLine(line);
                        line = "";
                    }
                    outfile.Close();
                }
            }
        }

        private void btnLoad_Click(object sender, EventArgs e)
        {
            OpenFileDialog ofd = new OpenFileDialog();
            ofd.Filter = "Text files|*.txt";
            DialogResult result = ofd.ShowDialog();
            string line = "";
            int rCounter = 0;
            int yCounter = 0;

            if (result == DialogResult.OK)
            {
                ResetBoard();
                using (StreamReader infile = new StreamReader(ofd.FileName))
                {
                    for (int j = 0; j < 6; j++)
                    {
                        line = infile.ReadLine();
                        for (int i = 0; i < 7; i++)
                        {
                            Engine.board[i, j] = line[i];
                            if (line[i] == 'r')
                            {
                                rCounter++;
                                Control[] pics = picBoard.Controls.Find("picToken" + i + j, false);
                                foreach (Control ctrl in pics)
                                {
                                    PictureBox pic = (PictureBox)ctrl;
                                    pic.Image = Resources.token_red;
                                }
                            }
                            else if (line[i] == 'y')
                            {
                                yCounter++;
                                Control[] pics = picBoard.Controls.Find("picToken" + i + j, false);
                                foreach (Control ctrl in pics)
                                {
                                    PictureBox pic = (PictureBox)ctrl;
                                    pic.Image = Resources.token_yellow;
                                }
                            }
                            else if (line[i] != 'n')
                            {
                                MessageBox.Show("Le format du fichier à charger est invalide!", "Format invalide");
                                return;
                            }
                        }
                    }
                }
                PlayersSettings();
                Engine.turn = rCounter;
                if (rCounter > yCounter)
                {
                    currentPlayer = player2;
                    lblCurrentPlayer.Text = "2 - Jaune";
                    if (Engine.gameType == PlayersType.HumanComputer)
                    {
                        Position posDeJeu = AI.Choisir(Engine.board, AI.profondeur, currentPlayer);
                        MettreJeton(posDeJeu);
                    }
                }
                else
                {
                    Engine.turn++;
                    currentPlayer = player1;
                    lblCurrentPlayer.Text = "1 - Rouge";
                }

                if(Engine.gameType == PlayersType.ComputerComputer)
                {
                    timerCompVSComp.Start();
                }

                lblTurn.Text = Engine.turn.ToString();
            }
        }

        private void MettreJeton(Position p_Pos)
        {
            char jetonAMettre = 'n';
            if (currentPlayer.IsRed)
            {
                jetonAMettre = 'r';
            }
            else
            {
                jetonAMettre = 'y';
            }
            Engine.board[p_Pos.colonne, p_Pos.ligne] = jetonAMettre;

            tokensOnBoard++;
            Control[] pics = picBoard.Controls.Find("picToken" + p_Pos.colonne + p_Pos.ligne, false);
            foreach (Control ctrl in pics)
            {
                PictureBox pic = (PictureBox)ctrl;
                if (currentPlayer.IsRed)
                {
                    pic.Image = Resources.token_red;
                }
                else
                {
                    pic.Image = Resources.token_yellow;
                }
                this.Refresh();
            }

            if (Engine.turn >= 4)
            {
                bool win = Engine.CheckWinState(currentPlayer, p_Pos.ligne, p_Pos.colonne);
                if (win)
                {
                    timerCompVSComp.Stop();
                    if (currentPlayer.PlayerNo == 1)
                    {
                        MessageBox.Show("Le joueur rouge a gagné la partie!", "Partie terminée!");
                        playerWon = true;
                    }
                    else
                    {
                        MessageBox.Show("Le joueur jaune a gagné la partie!", "Partie terminée!");
                    }
                        
                    ResetBoard();
                    if (Engine.gameType == PlayersType.ComputerComputer)
                    {
                        timerCompVSComp.Start();
                    }
                    return;
                }
            }

            if (currentPlayer.PlayerNo == 1)
            {
                currentPlayer = player2;
                lblCurrentPlayer.Text = "2 - Jaune";
            }
            else
            {
                currentPlayer = player1;
                lblCurrentPlayer.Text = "1 - Rouge";
                Engine.turn++;
                lblTurn.Text = Engine.turn.ToString();
            }

            if (tokensOnBoard == 42)
            {
                MessageBox.Show("La partie s'est terminée avec une égalité!\nAppuyez OK pour recommencer!", "Partie terminée!");
                ResetBoard();
            }
        }

        private void timerCompVSComp_Tick(object sender, EventArgs e)
        {
            Position posDeJeu = AI.Choisir(Engine.board, AI.profondeur, currentPlayer);
            MettreJeton(posDeJeu);
        }
    }
}
